'''
Created on 4 jan 2015

The goal is to take the flat project structure (in uVision 4) generated when exporting from mbed:

NXP LPC4088
    - src
        - file 1
        - file 2
        ...
        - file N
        
and, based on actual paths and some preferred paths, transform it into:

NXP LPC4088
    - Support/mbed-src
    - Support/mbed-rpc
    - Support/mbed-rtos
    - Support/Bios
    - Support/Display
    - Support/EthernetInterface
    - Support/FATFileSystem
    - Support/FileSystems
    - Support/HttpServer
    - Support/Memory
    - Support/USBDevice
    - Support/USBHost
    - Support
    - BasicGUI/Application
    - BasicGUI/lpc_swim
    - BasicGUI/SlideShow
    - src
    
All files under each folder are also sorted alphabetically (case insensitive). A couple of
imprtant h-files are also added as those are skipped by the mbed exporter.

@author: lindvalla
'''
import argparse
import re
import os

MODE_COPY_THE_REST   = 0
MODE_FIND_GROUPS     = 1
MODE_FIND_GROUP      = 2
MODE_FIND_GROUP_NAME = 3
MODE_FIND_FILES      = 4
MODE_FIND_FILE       = 5

FOLDERS = [
        [ "Support/mbed-src", "DMSupport/mbed-src" ],
        [ "Support/mbed-rpc", "DMSupport/mbed-rpc" ],
        [ "Support/mbed-rtos", "DMSupport/mbed-rtos" ],
        [ "Support/Bios" , "DMSupport/Bios"],
        [ "Support/Display" , "DMSupport/Display"],
        [ "Support/EthernetInterface", "DMSupport/EthernetInterface" ],
        [ "Support/FATFileSystem", "DMSupport/FATFileSystem", "DMSupport/DM_FATFileSystem" ],
        [ "Support/FileSystems", "DMSupport/FileSystems" ],
        [ "Support/HttpServer", "DMSupport/HttpServer", "DMSupport/DM_HttpServer" ],
        [ "Support/Memory", "DMSupport/Memory" ],
        [ "Support/USBDevice", "DMSupport/USBDevice" ],
        [ "Support/USBHost", "Support/LPC4088-USBHost", "DMSupport/DM_USBHost", "DMSupport/USBHost" ],
        [ "Support", "DMSupport" ],
        [ "BasicGUI/Application", "DMBasicGUI/Application" ],
        [ "BasicGUI/lpc_swim", "DMBasicGUI/lpc_swim" ],
		[ "BasicGUI/SlideShow", "DMBasicGUI/SlideShow" ],
		[ "emWin HAL", "DMemWin", "DMemWinGUI" ],
    ]

class AFile(object):
    
    def __init__(self, line):
        self.lines = []
        self.lines.append(line)
    
    def addLine(self, line):
        ts = line.strip()
        if ts.startswith("<FilePath>") and ts.endswith("</FilePath>"):
            self.path = re.compile("<FilePath>([^<]+)</FilePath>").split(line)[1]
            #print "  New file: ", self.path
        if ts.startswith("<FileName>") and ts.endswith("</FileName>"):
            self.name = re.compile("<FileName>([^<]+)</FileName>").split(line)[1]
        self.lines.append(line)


class Group(object):
    
    def __init__(self, line):
        if line.find("<GroupName>") >= 0:
            self.name = re.compile("<GroupName>([^<]+)</GroupName>").split(line)[1]
        else:
            self.name = line
        #print "New group: ", self.name
        self.files = []
        self.aliases = [self.name]   
    
    def addFile(self, line):
        self.files.append(AFile(line))

    def addLine(self, line):
        self.files[len(self.files)-1].addLine(line)
        
    def addAlias(self, alias):
        self.aliases.append(alias)
        
    def addCompleteFile(self, f):
        self.files.append(f)

    def matches(self, path):
        for alias in self.aliases:
            if path.startswith(alias):
                return True
        return False

class Parser(object):
    
    def __init__(self, fname):
        self.fin = open(fname, "r")
        self.fout = open(fname+".pyout", "w")
        self.mode = MODE_FIND_GROUPS
        self.groups = []
        
    def done(self):
        self.fin.close()
        self.fout.flush()
        self.fout.close()

    def insertHFile(self, group, fname, path, indent="            "):
        found = False;
        for f in group.files:
            if f.name == fname:
                found = True
                break
        if not found:
            group.addFile(indent + "<File>\n")
            group.addLine(indent + "  <FileName>{}</FileName>\n".format(fname))
            group.addLine(indent + "  <FileType>5</FileType>\n")
            group.addLine(indent + "  <FilePath>{}</FilePath>\n".format(path))
            group.addLine(indent + "</File>\n")    

    def insertConfigFile(self, group, indent="            "):
        self.insertHFile(group, "dm_board_config.h", "dm_board_config.h")

    def insertDisplayFiles(self, group, indent="            "):
        self.insertHFile(group, "Display.h", "DMSupport/Display/Display.h")
        self.insertHFile(group, "TouchPanel.h", "DMSupport/Display/TouchPanel.h")
    
    def reorganizeGroups(self):
        src = Group("src")
        
        newgroups = []
        for list in FOLDERS:
            g = Group(list[0])
            for a in list[1:len(list)]:
                g.addAlias(a)
            newgroups.append(g)
            
        for group in self.groups:
            for file in group.files:
                found = False
            
                for newg in newgroups:
                    if newg.matches(file.path):
                        #print "newg {} - {}".format(newg.name, file.path)
                        newg.addCompleteFile(file)
                        found = True
                        break
                if not found:
                    if file.path.find("/") >= 0:
                        g = Group(file.path[0:file.path.find("/")])
                        g.addCompleteFile(file)
                        newgroups.append(g)
                    else:
                        src.addCompleteFile(file)
            
        self.insertConfigFile(src)
        newgroups.append(src)  # the "src" folder should always be last
        
        for group in newgroups:
            if len(group.files) > 0:
                if group.name == "Support/Display":
                    self.insertDisplayFiles(group)
                
                self.fout.write("<Group>\n")
                self.fout.write("  <GroupName>{}</GroupName>\n".format(group.name))
                self.fout.write("  <Files>\n")
                #for f in group.files:
                #print "----"
                #print group.files
                #print "++++"
                #print sorted(group.files, key=lambda fobj: fobj.name.lower())
                #print "===="
                for f in sorted(group.files, key=lambda fobj: fobj.name.lower()):
                    for line in f.lines:
                        self.fout.write(line)
                self.fout.write("  </Files>\n")
                self.fout.write("</Group>\n")
                self.fout.flush()
    
    def process(self): 
        for line in self.fin.readlines():
            ts = line.strip()
            if self.mode == MODE_COPY_THE_REST:
                self.fout.write(line)
            elif self.mode == MODE_FIND_GROUPS:
                self.fout.write(line)
                if ts == "<Groups>":
                    self.mode = MODE_FIND_GROUP
            elif self.mode == MODE_FIND_GROUP:
                if ts == "<Group>":
                    self.mode = MODE_FIND_GROUP_NAME
                elif ts == "</Groups>":
                    self.mode = MODE_COPY_THE_REST
                    self.reorganizeGroups()
                    self.fout.write(line)
            elif self.mode == MODE_FIND_GROUP_NAME:
                if ts.startswith("<GroupName>") and ts.endswith("</GroupName>"):
                    self.mode = MODE_FIND_FILES                    
                    self.groups.append(Group(line))
            elif self.mode == MODE_FIND_FILES:
                if ts == "<Files>":
                    self.mode = MODE_FIND_FILE
            elif self.mode == MODE_FIND_FILE:
                if ts == "<File>":
                    self.groups[len(self.groups)-1].addFile(line)
                elif ts == "</Files>":
                    self.mode = MODE_FIND_GROUP
                elif ts != "":
                    self.groups[len(self.groups)-1].addLine(line)
                 


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("uvproj", help="uVision project file to reorganize")
    args = parser.parse_args()
    if os.path.isfile(args.uvproj):
        p = Parser(args.uvproj)
        p.process()
        p.done()
    else:
        print "Supplied argument '{}' is not a file!".format(args.uvproj)
